@@ -1,14 +1,22 @@
 """
-SQLObject 0.7.10
-----------------
+SQLObject
+---------
 
 :author: Ian Bicking <ianb@colorstudy.com>
 
 SQLObject is a object-relational mapper.  See SQLObject.html or
 SQLObject.txt for more.
 
-With the help by Oleg Broytmann <phd@phd.pp.ru> and many other contributors.
-See Authors.txt.
+Modified by:
+
+* Daniel Savard, Xsoli Inc <sqlobject xsoli.com> 7 Feb 2004
+  Added support for simple table inheritance.
+
+* Oleg Broytmann, SIA "ANK" <phd@phd.pp.ru> 3 Feb 2005
+  Split inheritance support into a number of separate modules and classes -
+  InheritableSQLObject at al.
+
+* And others...
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as
@@ -27,6 +35,7 @@ USA.
 """
 
 import threading
+import weakref
 import sqlbuilder
 import dbconnection
 import col
@@ -37,6 +46,7 @@ import joins
 import index
 import classregistry
 import declarative
+import events
 from sresults import SelectResults
 from formencode import schema, compound
 
@@ -210,7 +220,7 @@ class sqlmeta(object):
     __metaclass__ = declarative.DeclarativeMeta
 
     # These attributes shouldn't be shared with superclasses:
-    _unshared_attributes = ['table', 'idName', 'columns']
+    _unshared_attributes = ['table', 'idName', 'columns', 'childName']
 
     # These are internal bookkeeping attributes; the class-level
     # definition is a default for the instances, instances will
@@ -230,13 +240,24 @@ class sqlmeta(object):
     # will be true.  It's false by default:
     _perConnection = False
 
+    # Inheritance definitions:
+    parentClass = None # A reference to the parent class
+    childClasses = {} # References to child classes, keyed by childName
+    childName = None # Class name for inheritance child object creation
+
     def __classinit__(cls, new_attrs):
         for attr in cls._unshared_attributes:
             if not new_attrs.has_key(attr):
                 setattr(cls, attr, None)
+        declarative.setup_attributes(cls, new_attrs)
 
     def __init__(self, instance):
-        self.instance = instance
+        self.instance = weakref.proxy(instance)
+
+    def send(cls, signal, *args, **kw):
+        events.send(signal, cls.soClass, *args, **kw)
+
+    send = classmethod(send)
 
     def setClass(cls, soClass):
         cls.soClass = soClass
@@ -287,6 +308,9 @@ class sqlmeta(object):
     ########################################
 
     def addColumn(cls, columnDef, changeSchema=False, connection=None):
+        post_funcs = []
+        cls.send(events.AddColumnSignal, cls.soClass, connection,
+                 columnDef.name, columnDef, changeSchema, post_funcs)
         sqlmeta = cls
         soClass = cls.soClass
         del cls
@@ -299,8 +323,18 @@ class sqlmeta(object):
             "The class %s.%s already has a column %r (%r), you cannot "
             "add the column %r"
             % (soClass.__module__, soClass.__name__, name,
-               sqlmeta.columnDefinitions[name],
-               columnDef))
+               sqlmeta.columnDefinitions[name], columnDef))
+        # Collect columns from the parent classes to test
+        # if the column is not in a parent class
+        parent_columns = []
+        for base in soClass.__bases__:
+            if hasattr(base, "sqlmeta"):
+                parent_columns.extend(base.sqlmeta.columns.keys())
+        if hasattr(soClass, name):
+            assert  (name in parent_columns) or (name == "childName"), (
+                "The class %s.%s already has a variable or method %r, you cannot "
+                "add the column %r"
+                % (soClass.__module__, soClass.__name__, name, name))
         sqlmeta.columnDefinitions[name] = columnDef
         sqlmeta.columns[name] = column
         # A stable-ordered version of the list...
@@ -362,10 +396,10 @@ class sqlmeta(object):
         # SQLObject instance.
         if column.foreignKey:
 
-            # We go through the standard _SO_get_columnName
-            # deal, except chopping off the "ID" ending since
+            # We go through the standard _SO_get_columnName deal
             # we're giving the object, not the ID of the
             # object this time:
+            origName = column.origName
             if sqlmeta.cacheValues:
                 # self._SO_class_className is a reference
                 # to the class in question.
@@ -373,30 +407,22 @@ class sqlmeta(object):
             else:
                 # Same non-caching version as above.
                 getter = eval('lambda self: self._SO_foreignKey(self._SO_getValue(%s), self._SO_class_%s)' % (repr(name), column.foreignKey))
-            if column.origName.upper().endswith('ID'):
-                origName = column.origName[:-2]
-            else:
-                origName = column.origName
             setattr(soClass, rawGetterName(origName), getter)
 
             # And we set the _get_columnName version
-            # (sans ID ending)
-            if not hasattr(soClass, getterName(name)[:-2]):
-                setattr(soClass, getterName(name)[:-2], getter)
-                sqlmeta._plainForeignGetters[name[:-2]] = 1
+            if not hasattr(soClass, getterName(origName)):
+                setattr(soClass, getterName(origName), getter)
+                sqlmeta._plainForeignGetters[origName] = 1
 
             if not column.immutable:
                 # The setter just gets the ID of the object,
                 # and then sets the real column.
                 setter = eval('lambda self, val: setattr(self, %s, self._SO_getID(val))' % (repr(name)))
-                setattr(soClass, rawSetterName(name)[:-2], setter)
-                if not hasattr(soClass, setterName(name)[:-2]):
-                    setattr(soClass, setterName(name)[:-2], setter)
-                    sqlmeta._plainForeignSetters[name[:-2]] = 1
+                setattr(soClass, rawSetterName(origName), setter)
+                if not hasattr(soClass, setterName(origName)):
+                    setattr(soClass, setterName(origName), setter)
+                    sqlmeta._plainForeignSetters[origName] = 1
 
-            # We'll need to put in a real reference at
-            # some point.  See needSet at the top of the
-            # file for more on this.
             classregistry.registry(sqlmeta.registry).addClassCallback(
                 column.foreignKey,
                 lambda foreign, me, attr: setattr(me, attr, foreign),
@@ -413,6 +439,9 @@ class sqlmeta(object):
         if soClass._SO_finishedClassCreation:
             makeProperties(soClass)
 
+        for func in post_funcs:
+            func(soClass, column)
+
     addColumn = classmethod(addColumn)
 
     def addColumnsFromDatabase(sqlmeta, connection=None):
@@ -426,7 +455,8 @@ class sqlmeta(object):
 
     addColumnsFromDatabase = classmethod(addColumnsFromDatabase)
 
-    def delColumn(sqlmeta, column, changeSchema=False, connection=None):
+    def delColumn(cls, column, changeSchema=False, connection=None):
+        sqlmeta = cls
         soClass = sqlmeta.soClass
         if isinstance(column, str):
             column = sqlmeta.columns[column]
@@ -438,6 +468,9 @@ class sqlmeta(object):
             else:
                 raise IndexError(
                     "Column with definition %r not found" % column)
+        post_funcs = []
+        cls.send(events.DeleteColumnSignal, connection, column.name, column,
+                 post_funcs)
         name = column.name
         del sqlmeta.columns[name]
         del sqlmeta.columnDefinitions[name]
@@ -449,12 +482,12 @@ class sqlmeta(object):
         if sqlmeta._plainSetters.has_key(name):
             delattr(soClass, setterName(name))
         if column.foreignKey:
-            delattr(soClass, rawGetterName(name)[:-2])
-            if sqlmeta._plainForeignGetters.has_key(name[:-2]):
-                delattr(soClass, getterName(name)[:-2])
-            delattr(soClass, rawSetterName(name)[:-2])
-            if sqlmeta._plainForeignSetters.has_key(name[:-2]):
-                delattr(soClass, setterName(name)[:-2])
+            delattr(soClass, rawGetterName(soClass.sqlmeta.style.instanceIDAttrToAttr(name)))
+            if sqlmeta._plainForeignGetters.has_key(name):
+                delattr(soClass, getterName(name))
+            delattr(soClass, rawSetterName(soClass.sqlmeta.style.instanceIDAttrToAttr(name)))
+            if sqlmeta._plainForeignSetters.has_key(name):
+                delattr(soClass, setterName(name))
         if column.alternateMethodName:
             delattr(soClass, column.alternateMethodName)
 
@@ -465,6 +498,9 @@ class sqlmeta(object):
         if soClass._SO_finishedClassCreation:
             unmakeProperties(soClass)
 
+        for func in post_funcs:
+            func(soClass, column)
+
     delColumn = classmethod(delColumn)
 
     ########################################
@@ -566,8 +602,36 @@ class sqlmeta(object):
         cls.indexDefinitions.append(indexDef)
         index = indexDef.withClass(cls.soClass)
         cls.indexes.append(index)
+        setattr(cls.soClass, index.name, index)
     addIndex = classmethod(addIndex)
 
+    ########################################
+    ## Utility methods
+    ########################################
+
+    def asDict(self):
+        """
+        Return the object as a dictionary of columns to values.
+        """
+        result = {}
+        for key in self.columns:
+            result[key] = getattr(self.instance, key)
+        result['id'] = self.instance.id
+        return result
+
+    def expireAll(sqlmeta, connection=None):
+        """
+        Expire all instances of this class.
+        """
+        soClass = sqlmeta.soClass
+        connection = connection or soClass._connection
+        cache_set = connection.cache
+        cache_set.weakrefAll(soClass)
+        for item in cache_set.getAll(soClass):
+            item.expire()
+
+    expireAll = classmethod(expireAll)
+
 sqlhub = dbconnection.ConnectionHub()
 
 class _sqlmeta_attr(object):
@@ -650,10 +714,11 @@ class SQLObject(object):
     #DSM: The _inheritable attribute controls wheter the class can by
     #DSM: inherited 'logically' with a foreignKey and a back reference.
     _inheritable = False # Is this class inheritable?
-    _parentClass = None # A reference to the parent class
     _parent = None # A reference to the parent instance
-    _childClasses = {} # References to child classes
     childName = None # Children name (to be able to get a subclass)
+    # moved to sqlmeta in 0.8:
+    _parentClass = _sqlmeta_attr('parentClass', 2)
+    _childClasses = _sqlmeta_attr('childClasses', 2)
 
     # The law of Demeter: the class should not call another classes by name
     SelectResultsClass = SelectResults
@@ -706,44 +771,6 @@ class SQLObject(object):
             for j in new_attrs['_joins']:
                 implicitJoins.append(j)
 
-
-        #DSM: Need to know very soon if the class is a child of an
-        #DSM: inheritable class. If so, we keep a link to our parent class.
-        cls._childClasses = {}
-        for superclass in cls.__bases__:
-            if hasattr(superclass, '_inheritable') and superclass._inheritable and \
-                superclass.__name__ <> "InheritableSQLObject":
-                    cls._parentClass = superclass
-                    cls._parent = None
-                    superclass._childClasses[cls.__name__] = cls
-
-        #DSM: If this class is a child of a parent class, we need to do some
-        #DSM: attribute check and add a foreign key to the parent.
-        if cls._parentClass:
-            #DSM: First, look for invalid column name:
-            #DSM: reserved ones or same as a parent
-            parentCols = cls.sqlmeta.columns.keys()
-            for column in implicitColumns:
-                cname = column.name
-                if cname == 'childName':
-                    raise AttributeError, \
-                        "The column name 'childName' is reserved"
-                if cname in parentCols:
-                    raise AttributeError, "The column '%s' is already " \
-                        "defined in an inheritable parent" % cname
-            #DSM: Remove columns if inherited from an inheritable class
-            #DSM: as we don't want them.  All we want is a foreign key
-            #DSM: that points to our parent
-            cls.sqlmeta.columns = {}
-            cls.sqlmeta.columnList = []
-            cls.sqlmeta.columnDefinitions = {}
-        #DSM: If this is inheritable, add some default columns
-        #DSM: to be able to link to children
-        if hasattr(cls, '_inheritable') and cls._inheritable and \
-            cls.__name__ <> "InheritableSQLObject":
-            cls.sqlmeta.columnDefinitions['childName'] = col.StringCol(
-                name='childName', length=255, default=None)
-
         ######################################################
         # Set some attributes to their defaults, if necessary.
         # First we get the connection:
@@ -757,6 +784,10 @@ class SQLObject(object):
         if connection and not hasattr(cls, '_connection'):
             cls.setConnection(connection)
 
+        # Now the class is in an essentially OK-state, so we can
+        # set up any magic attributes:
+        declarative.setup_attributes(cls, new_attrs)
+
         # We have to check if there are columns in the inherited
         # _columns where the attribute has been set to None in this
         # class.  If so, then we need to remove that column from
@@ -792,17 +823,6 @@ class SQLObject(object):
         if not is_base:
             cls.q = sqlbuilder.SQLObjectTable(cls)
 
-            #DSM: If we are a child, get the q magic from the parent
-            currentClass = cls
-            while currentClass._parentClass:
-                currentClass = currentClass._parentClass
-                for column in currentClass.sqlmeta.columnDefinitions.values():
-                    if type(column) == col.ForeignKey: continue
-                    if column.name == 'childName':
-                        continue
-                    setattr(cls.q, column.name,
-                        getattr(currentClass.q, column.name))
-
         classregistry.registry(cls.sqlmeta.registry).addClass(cls)
 
     _style = _sqlmeta_attr('style', 2)
@@ -936,10 +956,6 @@ class SQLObject(object):
         # created, unlike __init__ which would be called
         # anytime the object was returned from cache.
         self.id = id
-        # We shadow the sqlmeta class with an instance of sqlmeta
-        # that points to us (our sqlmeta buddy object; where the
-        # sqlmeta class is our class's buddy class)
-        self.sqlmeta = self.__class__.sqlmeta(self)
         self._SO_writeLock = threading.Lock()
 
         # If no connection was given, we'll inherit the class
@@ -1041,6 +1057,12 @@ class SQLObject(object):
         # in the database, and we can't insert it until all
         # the parts are set.  So we just keep them in a
         # dictionary until later:
+        d = {name: value}
+        if not self.sqlmeta._creating:
+            self.sqlmeta.send(events.RowUpdateSignal, self, d)
+        if len(d) != 1 or name not in d:
+            return self.set(**d)
+        value = d[name]
         if from_python:
             dbValue = from_python(value, self._SO_validatorState)
         else:
@@ -1061,6 +1083,8 @@ class SQLObject(object):
             setattr(self, instanceName(name), value)
 
     def set(self, **kw):
+        if not self.sqlmeta._creating:
+            self.sqlmeta.send(events.RowUpdateSignal, self, kw)
         # set() is used to update multiple values at once,
         # potentially with one SQL statement if possible.
 
@@ -1177,17 +1201,25 @@ class SQLObject(object):
             return joinClass.get(id)
 
     def __init__(self, **kw):
+        # We shadow the sqlmeta class with an instance of sqlmeta
+        # that points to us (our sqlmeta buddy object; where the
+        # sqlmeta class is our class's buddy class)
+        self.sqlmeta = self.__class__.sqlmeta(self)
         # The get() classmethod/constructor uses a magic keyword
         # argument when it wants an empty object, fetched from the
         # database.  So we have nothing more to do in that case:
         if kw.has_key('_SO_fetch_no_create'):
             return
 
+        post_funcs = []
+        self.sqlmeta.send(events.RowCreateSignal, kw, post_funcs)
+
         # Pass the connection object along if we were given one.
         if kw.has_key('connection'):
             self._connection = kw['connection']
             self.sqlmeta._perConnection = True
             del kw['connection']
+
         self._SO_writeLock = threading.Lock()
 
         if kw.has_key('id'):
@@ -1197,6 +1229,8 @@ class SQLObject(object):
             id = None
 
         self._create(id, **kw)
+        for func in post_funcs:
+            func(self)
 
     def _create(self, id, **kw):
 
@@ -1218,6 +1252,7 @@ class SQLObject(object):
                     raise TypeError, "%s() did not get expected keyword argument %s" % (self.__class__.__name__, column.name)
                 # Otherwise we put it in as though they did pass
                 # that keyword:
+
                 kw[column.name] = default
 
         self.set(**kw)
@@ -1251,6 +1286,10 @@ class SQLObject(object):
         cache = self._connection.cache
         cache.created(id, self.__class__, self)
         self._init(id)
+        post_funcs = []
+        kw = dict([('class',self.__class__),('id',id)])
+        self.sqlmeta.send(events.RowCreatedSignal, kw, post_funcs)
+
 
     def _SO_getID(self, obj):
         return getID(obj)
@@ -1268,10 +1307,17 @@ class SQLObject(object):
             value), None
     _findAlternateID = classmethod(_findAlternateID)
 
-    def _SO_fetchAlternateID(cls, name, dbName, value, connection=None):
+    def _SO_fetchAlternateID(cls, name, dbName, value, connection=None, idxName=None):
         result, obj = cls._findAlternateID(name, dbName, value, connection)
         if not result:
-            raise SQLObjectNotFound, "The %s by alternateID %s=%s does not exist" % (cls.__name__, name, repr(value))
+            if idxName is None:
+                raise SQLObjectNotFound, "The %s by alternateID %s = %s does not exist" % (cls.__name__, name, repr(value))
+            else:
+                names = []
+                for i in xrange(len(name)):
+                    names.append("%s = %s" % (name[i], repr(value[i])))
+                names = ', '.join(names)
+                raise SQLObjectNotFound, "The %s by unique index %s(%s) does not exist" % (cls.__name__, idxName, names)
         if obj:
            return obj
         if connection:
@@ -1289,7 +1335,7 @@ class SQLObject(object):
                orderBy=NoDefault, limit=None,
                lazyColumns=False, reversed=False,
                distinct=False, connection=None,
-               join=None):
+               join=None, forUpdate=False):
         return cls.SelectResultsClass(cls, clause,
                              clauseTables=clauseTables,
                              orderBy=orderBy,
@@ -1298,7 +1344,7 @@ class SQLObject(object):
                              reversed=reversed,
                              distinct=distinct,
                              connection=connection,
-                             join=join)
+                             join=join, forUpdate=forUpdate)
     select = classmethod(select)
 
     def selectBy(cls, connection=None, **kw):
@@ -1319,35 +1365,59 @@ class SQLObject(object):
         conn = connection or cls._connection
         if ifExists and not conn.tableExists(cls.sqlmeta.table):
             return
+        extra_sql = []
+        post_funcs = []
+        cls.sqlmeta.send(events.DropTableSignal, cls, connection,
+                         extra_sql, post_funcs)
         conn.dropTable(cls.sqlmeta.table, cascade)
         if dropJoinTables:
             cls.dropJoinTables(ifExists=ifExists, connection=conn)
+        for sql in extra_sql:
+            connection.query(sql)
+        for func in post_funcs:
+            func(cls, conn)
     dropTable = classmethod(dropTable)
 
     def createTable(cls, ifNotExists=False, createJoinTables=True,
-                    createIndexes=True,
+                    createIndexes=True, applyConstraints=True,
                     connection=None):
         conn = connection or cls._connection
         if ifNotExists and conn.tableExists(cls.sqlmeta.table):
             return
-        conn.createTable(cls)
+        extra_sql = []
+        post_funcs = []
+        cls.sqlmeta.send(events.CreateTableSignal, cls, connection,
+                         extra_sql, post_funcs)
+        constraints = conn.createTable(cls)
+        if applyConstraints:
+            for constraint in constraints:
+                conn.query(constraint)
+        else:
+            extra_sql.extend(constraints)
         if createJoinTables:
             cls.createJoinTables(ifNotExists=ifNotExists,
                                  connection=conn)
         if createIndexes:
             cls.createIndexes(ifNotExists=ifNotExists,
                               connection=conn)
+        for func in post_funcs:
+            func(cls, conn)
+        return extra_sql
     createTable = classmethod(createTable)
 
-    def createTableSQL(cls, createJoinTables=True, connection=None,
-                       createIndexes=True):
+    def createTableSQL(cls, createJoinTables=True, createIndexes=True,
+                       connection=None):
         conn = connection or cls._connection
-        sql = conn.createTableSQL(cls)
+        sql, constraints = conn.createTableSQL(cls)
         if createJoinTables:
-            sql += ';\n' + cls.createJoinTablesSQL(connection=conn)
+            join_sql = cls.createJoinTablesSQL(connection=conn)
+            if join_sql:
+                sql += ';\n' + join_sql
         if createIndexes:
-            sql += ';\n' + cls.createIndexesSQL(connection=conn)
-        return sql
+            index_sql = cls.createIndexesSQL(connection=conn)
+            if index_sql:
+                sql += ';\n' + index_sql
+        return sql, constraints
     createTableSQL = classmethod(createTableSQL)
 
     def createJoinTables(cls, ifNotExists=False, connection=None):
@@ -1431,6 +1501,7 @@ class SQLObject(object):
     clearTable = classmethod(clearTable)
 
     def destroySelf(self):
+        self.sqlmeta.send(events.RowDestroySignal, self)
         # Kills this object.  Kills it dead!
 
         klass = self.__class__
@@ -1499,6 +1570,9 @@ class SQLObject(object):
     delete = classmethod(delete)
 
     def __repr__(self):
+        if not hasattr(self, 'id'):
+            # Object initialization not finished.  No attributes can be read.
+            return '<%s (not initialized)>' % self.__class__.__name__
         return '<%s %r %s>' \
                % (self.__class__.__name__,
                   self.id,
@@ -1551,7 +1625,7 @@ def instanceName(name):
 class SQLObjectState(object):
 
     def __init__(self, soObject):
-        self.soObject = soObject
+        self.soObject = weakref.proxy(soObject)
         self.protocol = 'sql'
 
 
