@@ -57,6 +57,14 @@ class MySQLConnection(DBAPI):
     connectionFromURI = classmethod(connectionFromURI)
 
     def makeConnection(self):
+        dbEncoding = self.dbEncoding
+        if dbEncoding:
+            from MySQLdb.connections import Connection
+            if not hasattr(Connection, 'set_character_set'):
+                # monkeypatch pre MySQLdb 1.2.1
+                def character_set_name(self):
+                    return dbEncoding + '_' + dbEncoding
+                Connection.character_set_name = character_set_name
         try:
             conn = self.module.connect(host=self.host, port=self.port,
                 db=self.db, user=self.user, passwd=self.password, **self.kw)
@@ -71,6 +79,13 @@ class MySQLConnection(DBAPI):
         if hasattr(conn, 'autocommit'):
             conn.autocommit(bool(self.autoCommit))
 
+        if self.dbEncoding:
+            if hasattr(conn, 'set_character_set'): # MySQLdb 1.2.1 and later
+                conn.set_character_set(self.dbEncoding)
+            else: # pre MySQLdb 1.2.1
+                # works along with monkeypatching code above
+                conn.query("SET NAMES %s" % self.dbEncoding)
+
         return conn
 
     def _setAutoCommit(self, conn, auto):
@@ -92,13 +107,11 @@ class MySQLConnection(DBAPI):
         # done by calling ping(True) on the connection.
         for count in range(3):
             try:
-                if self.need_unicode:
-                    # For MysqlDB 1.2.1 and later, we go
-                    # encoding->unicode->charset (in the mysql db)
-                    myquery = unicode(query, self.encoding)
-                    return cursor.execute(myquery)
-                else:
-                    return cursor.execute(query)
+                # For MySQLdb 1.2.1 and later, we go
+                # encoding->unicode->charset (in the mysql db)
+                if self.need_unicode and not isinstance(query, unicode):
+                    query = unicode(query, self.encoding)
+                return cursor.execute(query)
             except MySQLdb.OperationalError, e:
                 if e.args[0] in (MySQLdb.constants.CR.SERVER_GONE_ERROR, MySQLdb.constants.CR.SERVER_LOST):
                     if count == 2:
